;ASSEMBLY SOURCE FOR BT.OBJ
;POWER BASIC 3.0(c) INTERFACE FOR BTRIEVE/N

EXTRN GETSTRLOC:  FAR           ;POWER BASIC FUNCTION
;
VAR_ID          EQU     06176H  ;VARIABLE data length ID
BTR_ERR         EQU     20      ;Return status when Btrieve not initialized
POS_LEN_ERR     EQU     23      ;Invalid position length status
;
BTR_VECTOR      EQU     07BH * 4
MULTI_FUNCTION  EQU     0ABH

PARM_OFF     =   6             ;4 Byte overhead(return address) plus push bp
OPCODE_PTR   =  20 +PARM_OFF
POSBLOCK_PTR =  16 +PARM_OFF
DATABUFF_PTR =  12 +PARM_OFF
LENGTH_PTR   =  8  +PARM_OFF
KEYBUF_PTR   =  4  +PARM_OFF
KEYNO_PTR    =  0  +PARM_OFF


data       segment byte
      assume ds:data
DB_OFF     DW   ?
DB_SEG     DW   ?
DB_LENGTH  DW   ?
POS_OFF    DW   ?
POS_SEG    DW   ?
FCB_OFF    DW   ?
FCB_SEG    DW   ?
OPCODE     DW   ?
KEYBUF_OFF DW   ?
KEYBUF_SEG DW   ?
KEYLEN     DB   ?
KEYNO      DB   ?
STAT_OFF   DW   ?
STAT_SEG   DW   ?
LANG_ID    DW   ?

BASIC_VEC  DD   0
PROCID     DW   0
MULTI      DW   0
VSET       DW   0
STAT       DW   ?       ;Btrieve returns status here.
IS_STRING  DW   0
data    ends



code        segment byte
    assume cs:code

BTRV    Proc far
	public BTRV
	push bp
	mov bp,sp
	push si
	push di
	push ds
	;SAVE BASIC CRITICAL ERROR HANDLER

	xor bx,bx
	mov es,bx                       ;es =>absolute 0
	mov cx,word ptr es:90H
	mov word ptr BASIC_VEC,cx
	mov cx,word ptr es:92H
	mov word ptr BASIC_VEC +2,cx
	mov cx,word ptr es:51AH
	mov es:90H,cx
	mov cx,word ptr es:51CH
	mov es:92H,cx
	;IS BTRIEVE LOADED?
	push ds
	sub bx,bx
	mov ds,bx
	cmp word ptr BTR_VECTOR[bx],33H
	pop ds
	je CHECK_DOS_MULTI                 ;yes
	jmp BTRIEVE_NOT_LOADED
CHECK_DOS_MULTI:
	cmp VSET,0
	jne START
	inc VSET
	MOV AX,3000H                     ;get dos version
	INT 21H
	cmp AL,3
	jb START                          ;dos 3 or higher?
	mov ax,MULTI_FUNCTION *256        ;is bmulti loaded?
	int 02FH
	cmp al,'M'
	jne START
	inc MULTI


START:
	;check opcode to determine data type
	mov IS_STRING,0
	les di,[bp+OPCODE_PTR]
	mov ax,word ptr es:[di]
	cmp ax,1000               ;is data a string?
	jb TYPED_VAR              ;no
	mov IS_STRING,1
	sub ax,1000
	mov word ptr es:[di],ax
	les di,[bp+DATABUFF_PTR]
	push word ptr es:[di]
	call GETSTRLOC
	mov DB_OFF,ax
	mov DB_SEG,dx
	mov DB_LENGTH,cx
TYPED_VAR:
	;FIRST PARAMETER   - OPCODE
	;Get Function  put it in  OPCODE
	les di,[bp+OPCODE_PTR]
	mov ax,es:[di]
	mov OPCODE,ax
	;SECOND AND THIRD PARAMETER  - POSITION BLOCK
	;Get the POSITION BLOCK   90 byte storage area  POS
	;FCB                      38 byte storage area  FCB
	;NOTE BTRIEVE expects the 90 byte block then the 38 byte block
	;we are storing the 38 byte block first in the passed string.

	les di,[bp+POSBLOCK_PTR]       ;pointer  to handle
	push word ptr es:[di]            ;and push it
	;check for length         ;
	call GETSTRLOC            ;return seg in dx
	cmp cx,128                ;length in cx
	jne POSBLOCK_TOO_SHORT
	mov FCB_OFF,ax            ;offset in ax
	mov FCB_SEG,dx
	add ax,38
	mov POS_OFF,ax            ;90 byte area
	mov POS_SEG,dx
	jmp POSBLOCK_OK
POSBLOCK_TOO_SHORT:
	mov STAT,POS_LEN_ERR
	jmp EARLY_OUT
POSBLOCK_OK:
	cmp IS_STRING,0
	jne DB_IS_STRING
	;FOURTH PARAMETER - DATA BUFFER
	; now the databuffer
	les di,[bp+DATABUFF_PTR]
	mov DB_OFF,di
	mov DB_SEG,es
	;FIFTH PARAMETER - DATA BUFFER LENGTH
	les di,[bp+LENGTH_PTR]
	mov ax,word ptr es:[di]
	mov DB_LENGTH,ax
DB_IS_STRING:
	;SIXTH & SEVENTH PARAMETER  - KEY BUFFER  & KEY BUFFER LENGTH
	les di,[bp+KEYBUF_PTR]       ;pointer  to handle
	push word ptr es:[di]        ;and push it
	call GETSTRLOC               ;return seg in dx
	mov KEYBUF_OFF,ax            ;offset in ax
	mov KEYBUF_SEG,dx
	mov KEYLEN,cl                ;length in cx

	;EIGHTH PARAMETER  - Key no
	les di,[bp+KEYNO_PTR]
	mov al,es:[di]
	mov KEYNO,al
	;NINETH PARAMATER   - STATUS RETURNED
	mov dx, OFFSET STAT
	mov STAT_OFF,dx
	mov STAT_SEG,ds
	;TENTH PARAMETER
	mov LANG_ID,VAR_ID
	MOV DX,OFFSET DB_OFF   ;Btrieve wants DS:DX to point to PARAM BLOCK
	;MULTI CALL
	cmp MULTI,0
	je NOT_MULTI
MAKE_CALL:
	xor ax,ax
	inc ax
	mov bx,PROCID
	or bx,bx
	je CALL_MULTI
	inc ax
CALL_MULTI:
	mov ah,MULTI_FUNCTION
	int 02FH
	cmp al,0
	je DONE_CALL
	mov ax,200H
	int 7FH
	jmp MAKE_CALL
DONE_CALL:
	cmp PROCID,bx
	jne NORMAL_EXIT
	mov PROCID,bx
	jmp NORMAL_EXIT
NOT_MULTI:
	INT 7BH
	JMP NORMAL_EXIT

BTRIEVE_NOT_LOADED:
	mov STAT,BTR_ERR
	JMP EARLY_OUT

NORMAL_EXIT:
	les di,[bp+LENGTH_PTR];
	mov ax,word ptr DB_LENGTH        ;
	mov word ptr es:[di],ax          ;copy length back
EARLY_OUT:
	xor bx,bx
	mov es,bx
	mov cx,word ptr BASIC_VEC
	mov es:90H,cx                   ;Restore basic's Error handler
	mov cx,word ptr BASIC_VEC+2
	mov es:92H,cx

	mov AX,stat     ;return code for PB function
	pop ds
	pop di
	pop si
	pop bp
	retf 24


BTRV   endp
code   ends


    end
