/* btrieve.h -- header file for btrieve */

#define B_OPEN              0
#define B_CLOSE             1
#define B_INSERT            2
#define B_UPDATE            3
#define B_DELETE            4
#define B_GETEQ             5
#define B_GETNX             6
#define B_GETPR             7
#define B_GETGREATER        8
#define B_GETGE             9
#define B_GETLESS          10
#define B_GETLE            11
#define B_GETLW            12
#define B_GETHI            13
#define B_CREATE           14
#define B_STAT             15
#define B_EXTEND           16
#define B_SETDIR           17
#define B_GETDIR           18
#define B_BEGIN            19
#define B_END              20
#define B_ABORT            21
#define B_GETPOS           22
#define B_GETDIRECT        23
#define B_STEP_NEXT        24
#define B_STOP             25
#define B_VERSION          26
#define B_UNLOCK           27
#define B_RESET            28
#define B_SET_OWNER        29
#define B_CLR_OWNER        30
#define B_CREATE_INDEX     31
#define B_DROP_INDEX       32
#define B_STEP_FIRST       33
#define B_STEP_LAST        34
#define B_STEP_PREVIOUS    35
#define B_GETNX_EXT        36
#define B_GETPR_EXT        37
#define B_STEP_NEXT_EXT    38
#define B_STEP_PREV_EXT    39
#define B_INSERT_EXT       40

#define GET_KEY            50

#define EOF_ERR             9

#define WAIT_LOCK	            100
#define NOWAIT_LOCK	         200
#define MULTIPLE_WAIT_LOCK    300
#define MULTIPLE_NOWAIT_LOCK  400

extern int init_btrv(unsigned int); /* pass the max btrieve record size */
extern int exit_btrv(void);
extern int BTRV(int, char *, void *, int *, void *, int);

/* FILE FLAGS */

#define VARIABLE_LENGTH    1
#define BLANK_COMPRS       2
#define PREALLOCATION      4
#define COMPRESSION        8
#define KEY_ONLY          16
#define TEN_PERCENT       64
#define TWENTY_PERCENT   128
#define THIRTY_PERCENT  192


/* EXTENDED KEY TYPES */

#define B_STRING             0
#define B_INTEGER            1
#define B_FLOAT              2
#define B_DATE               3
#define B_TIME               4
#define B_DECIMAL            5
#define B_MONEY              6
#define B_LOGICAL            7
#define B_NUMERIC            8
#define B_BFLOAT             9
#define B_LSTRING            10
#define B_ZSTRING            11
#define B_UNSIGNED_BINARY    14
#define B_AUTOINCREMENT      15


/* KEY FLAGS */

#define B_DUP          1
#define B_MOD          2
#define B_BIN          4
#define B_NULL_VALUE   8
#define B_SEGMENTED    16
#define B_ALT_SEQ      32
#define B_DESCENDING   64
#define B_SUPPLEMENTAL 128
#define B_EXTENDED     256
#define B_MANUAL       512

typedef struct KEY_SPEC {

   int   KEY_POS;
   int   KEY_LEN;
   int   KEY_FLAG;
   long  NOT_USED;
   char  KEY_TYPE;
   char  NULL_VALUE;
   char  RESERVED[4];

};

typedef struct FIL_SPEC {

   int   REC_LEN;
   int   PAGE_SIZ;
   int   NDX_CNT;
   long  REC_TOT;
   int   FILE_FLAGS;
   int   RESERVED;
   int   ALLOCATION;
   struct  KEY_SPEC KEY_BUF[24];
   char  AlternateSequence[265];

}BTRIEVE_FILE_BUFFER;

