/****************************************************************************
 *																			*
 *	O B J E C T   O R I E N T E D   D A T A B A S E							*
 *																			*
 *	This is an object oriented extension to the BTrieve record management	*
 * system that allows for tables and database manipulation in a C++			*
 * environment.  The library supports the following classes:				*
 *																			*
 *	Database	Controls the tables and fields.  Has methods to perform		*
 *				the opening and createion of database tables as well as		*
 *				general table controls.										*
 *																			*
 *  Table		Table description.  Has fields associated with it.  Is a	*
 *				pure virtual class and must be redefined to create the		*
 *				database tables.  Has methods for record manuipulation		*
 *				without keys and other functions.  To support the extended	*
 *				string type, a lvar field is automatically placed at the	*
 *				end of any table that has an aField whose size is larger	*
 *				than 255 bytes.  The table then allocates free space from 	*
 *				this area.  Free space compaction and allocation is done	*
 *				automatically.  The field is named Ox$<table>$ and field	*
 *				names of this should not be used.  This also limits the		*
 *				table name for aField tables to 16 characters.  If all		*
 *				aFields are less than the index portion of the aField then	*
 *				the free space will not be used.							*
 *																			*
 *	Field		Field base class.  This class is re-defined to handle		*
 *				specific field types.  The fldInfo structure must be filled	*
 *				used to identify the fields.  When a table is opened the	*
 *				data dictionary is used to locate a field.					*
 *				The following built-in types are defined:					*
 *																			*
 *					aField	Alpha numeric field.  Can be up to 32768 bytes	*
 *							long.  The header record can be up to 255 bytes	*
 *							long and can be indexed.  There is no SQL or	*
 *							Xtrieve equivalent.  To Xtrieve this looks like	*
 *							string type.  The remaining bytes are stored in	*
 *							the variable length part of the record.  Note	*
 *							that all strings are zero terminated strings	*
 *                                                                          *
 *					cField	Character field.  A 1 byte character field.		*
 *							can be used to store flags or small integers.	*
 *                                                                          *
 *					dField	Date field.  Stores the date as d-m-y in julian	*
 *							format.  this allows for date+days.  There are	*
 *							operators for the following:					*
 *																			*
 *								day( long )		returns day as 1..31		*
 *								month( long )	returns month as 1..12		*
 *								year( long )	returns year as 1..65535	*
 *								dtw( long )		returns day of week			*
 *								dim(long, int)	increment month by months	*
 *					                                                        *
 *					fField	Floating point.  4 Byte floating point.  Note	*
 *							because of the C convention of returning a		*
 *							double the field will always return double,		*
 *							not float.										*
 *																			*
 *					lField	Long number.  4 byte integer field.				*
 *																			*
 *					nField	Double floating point.  8 Byte floating point	*
 *																			*
 *					rField	Record sequence number.  An autoincrement field	*
 *							that is used as a sequence or identifier on all	*
 *							fields.											*
 *                                                                          *
 *					sField	Short integer.  A 2 byte integer field.			*
 *                                                                          *
 *					tField	Time field.  Stored as a long.  Has 4 byte		*
 *							fields. stored as:								*
 *								struct time {								*
 *									unsigned char hundredths;				*
 *  								unsigned char seconds;					*
 *									unsigned char minutes;					*
 *									unsigned char hours;					*
 *									};										*
 *							member functions are provided to return int's	*
 *							for each of these.								*
 *                                                                          *
 *				Other field types can be created by deriving them from the	*
 *				field base class.  To create a field the fldInfo sructure	*
 *				must be filled in on a call to create.						*
 *																			*
 *				The scalar type fields have an operator []() defined to 	*
 *				allow for sub-scripted fields.  These are stored as 		*
 *				contiguous fields in the database.							*
 *                                                                          *
 *	Key			All keys are defined using this base class.  Flags are		*
 *				passed to create the key.									*
 *			                                                                *
 ****************************************************************************/

// Datebase sepcific includes

#ifndef __DBTYPES_H
#define __DBTYPES_H

#ifdef TRUE
#undef TRUE
#endif

#ifdef FALSE
#undef FALSE
#endif

extern int GetOODBVersion();			// returns the version number

const int TRUE = 1;							// true value
const int FALSE = 0;						// false value
const int DB_YIELD = 2;						// called at the base of loops
const int DB_SETRECORDS = 1;				// set record counter
const int DB_RECORD = 0;					// increment record counter

#include <oodb\btrtype.h>

#endif