// This is the database dictionary management class

#include <_defs.h>
#include <cstring.h>
#include <oodb\dbtypes.h>
#include <oodb\btrvcall.h>
#include <oodb\datadict.h>

const int MAXNAMELEN = 20;					// database names
const int MAXPATHLEN = 64;					// database path lengths

extern void _FAR dbError(int);				// function to globally handle database
											// errors

// this is the upper case collating sequence used for keys
//
const char altSeq[] = {
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
	0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
	0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F,
	 ' ',  '!', '\"',  '#',  '$',  '%',  '&', '\'',
	 '(',  ')',  '*',  '+',  ',',  '-',  '.',  '/',
	 '0',  '1',  '2',  '3',  '4',  '5',  '6',  '7',
	 '8',  '9',  ':',  ';',  '<',  '=',  '>',  '?',
	 '@',  'A',  'B',  'C',  'D',  'E',  'F',  'G',
	 'H',  'I',  'J',  'K',  'L',  'M',  'N',  'O',
	 'P',  'Q',  'R',  'S',  'T',  'U',  'V',  'W',
	 'X',  'Y',  'Z',  '[', '\\',  ']',  '^',  '_',
	 '`',  'A',  'B',  'C',  'D',  'E',  'F',  'G',
	 'H',  'I',  'J',  'K',  'L',  'M',  'N',  'O',
	 'P',  'Q',  'R',  'S',  'T',  'U',  'V',  'W',
	 'X',  'Y',  'Z',  '{',  '|',  '}',  '~', 0x7F,
	0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
	0x88, 0x89, 0x8A, 0x8B, 0x8C, 0x8D, 0x8E, 0x8F,
	0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97,
	0x98, 0x99, 0x9A, 0x9B, 0x9C, 0x9D, 0x9E, 0x9F,
	0xA0, 0xA1, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6, 0xA7,
	0xA8, 0xA9, 0xAA, 0xAB, 0xAC, 0xAD, 0xAE, 0xAF,
	0xB0, 0xB1, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6, 0xB7,
	0xB8, 0xB9, 0xBA, 0xBB, 0xBC, 0xBD, 0xBE, 0xBF,
	0xC0, 0xC1, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7,
	0xC8, 0xC9, 0xCA, 0xCB, 0xCC, 0xCD, 0xCE, 0xCF,
	0xD0, 0xD1, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6, 0xD7,
	0xD8, 0xD9, 0xDA, 0xDB, 0xDC, 0xDD, 0xDE, 0xDF,
	0xE0, 0xE1, 0xE2, 0xE3, 0xE4, 0xE5, 0xE6, 0xE7,
	0xE8, 0xE9, 0xEA, 0xEB, 0xEC, 0xED, 0xEE, 0xEF,
	0xF0, 0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7,
	0xF8, 0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, 0xFF };

struct dictFile {				// format of the data dictionary FILE.DDF table
	short	Xf_Id;							// File ID assigned 1..n AUTOINC index type
	char	Xf_Name[ MAXNAMELEN ];			// FileName assigned
	char	Xf_Loc[ MAXPATHLEN ];			// path name to the table
	short	Xf_Flags;						// flags assigned
};

struct cfFile {					// create structure
	fileSpec	file;
	keySpec		Id;				// Xf$Id Key
	keySpec		Name;			//Xf$Name Key
	altCol		Col;			// upper case collating sequence

    cfFile();
};

cfFile::cfFile() :
	file(sizeof(dictFile), 512, 2, 0),
	Id(0, sizeof(short), Xi_Extended | Xi_Binary, btAutoinc),
	Name(2, MAXNAMELEN, Xi_Modifiable | Xi_AltCollateSeq),
	Col("FileUppr")
	{
	}

class TddFile {
public:
	TddFile()
		{ memset(this, 0, sizeof( TddFile )); }
	TddFile(const char *path)
		{
		memset(this, 0, sizeof( TddFile ));
		strncpy(Xf_Path, path, MAXPATHLEN);
		strncat(Xf_Path, "\\FILE.DDF", MAXPATHLEN);
		}
	~TddFile()
		{
		}

	void setPath(const char *path)
		{
		strncpy(Xf_Path, path, MAXPATHLEN);
		strncat(Xf_Path, "\\FILE.DDF", MAXPATHLEN);
		}

	int open()						// open the FILE.DDF table
		{
		int Result;
		if( isOpen )
        	return btInvalidOperation;

		if( (Result = btOpen(posBlk, Xf_Path))== 0)		// just perform an open
			isOpen = TRUE;

		return Result;
		}

	int close()
	   {
	   isOpen = FALSE;
	   return btClose( posBlk );
	   }	

	short getId()
		{ return record.Xf_Id; }

	int dropFile(short fileId);

    int findFile(const char *fileName);

	int create();						// create a new FILE.DDF table

	int findId(short Id);				// get record where id == id

	const char *getLoc() const
		{ return (const char *) &record.Xf_Loc[0]; }

	int createSchema(RTdd p);			// sets the database schema

	friend int addTable(RTdd dict, const char *name, const char *path, short flags);

	// set the function to call at the bottom of loops.  allows
	// windows to multi-task when the system is busy
	void setLoop(void *obj, void (*func)(void *, int, long))
		{ object = obj; stFunc = func; }
	void yield()
    	{ if(stFunc) stFunc(object, DB_YIELD, 0); }

private:
	int		isOpen;							// true if open
	char 	Xf_Path[MAXPATHLEN+1];			// path name of the FILE.DDF table
	struct dictFile record;					// record read
    char	keyBuf[MAXNAMELEN];				// a key buffer
	char 	posBlk[ 128 ];					// Btrieve position block
	unsigned short	len;						// length of data read/written
	void (*stFunc)(void *, int, long);
    void *object;
};

struct dictField {				// format of the datadictionary FIELD.DDF table
	short	Xe_Id;							// Field Id assigned, 1..n AUTOINC type
	short	Xe_File;						// ID to file that this field belongs
	char	Xe_Name[ MAXNAMELEN ];			// Field Name
	char	Xe_DataType;					// type of data see btrieve_String etc. 0xFF == suplemental index
	short   Xe_Offset;						// field offset in file	index # if supplemental
	short	Xe_Size;						// field size in bytes
	char	Xe_Dec;							// field decimal places
	short	Xe_Flags;						// field flags
};

struct cfField {					// create structure
	fileSpec	file;
	keySpec		Id;					// Xf$Id Key
	keySpec		File;				//Xf$File Key
	keySpec		Name;				//Xf$Name Key
	keySpec		fldFile;			// Xf$File Key
	keySpec		uName;				// Xf$Name key (unique)
	altCol		Col;				// upper case collating sequence

    cfField();
};

cfField::cfField() :
	file(sizeof(dictField), 512, 4, 0),
	Id(0, sizeof(short), Xi_Extended | Xi_Binary, btAutoinc),
	File(2, sizeof(short), Xi_Binary | Xi_Duplicates),
	Name(4, MAXNAMELEN, Xi_Modifiable | Xi_Duplicates | Xi_AltCollateSeq),
	fldFile(2, sizeof(short), Xi_Modifiable | Xi_Binary | Xi_Segment),
	uName(4, MAXNAMELEN, Xi_Modifiable | Xi_AltCollateSeq),
	Col("UPPER   ")
	{
	}

class TddField {
public:
	TddField()
		{ memset(this, 0, sizeof( TddField )); }
	TddField(const char *path)
		{
		memset(this, 0, sizeof( TddField ));
		strncpy(Xe_Path, path, MAXPATHLEN);
		strncat(Xe_Path, "\\FIELD.DDF", MAXPATHLEN);
		}
	~TddField()
		{
		}

	void setPath(const char *path)
		{
		strncpy(Xe_Path, path, MAXPATHLEN);
		strncat(Xe_Path, "\\FIELD.DDF", MAXPATHLEN);
		}

	int open()						// open the FIELD.DDF table
		{
		int Result;
		if( isOpen )
        	return btInvalidOperation;

		if( (Result = btOpen(posBlk, Xe_Path))== 0)		// just perform an open
			isOpen = TRUE;

		return Result;
		}

	int close()
	   {
       isOpen = FALSE;
	   return btClose( posBlk );
	   }	

	short getId()
		{ return record.Xe_Id; }

	int dropField(short fieldId);					// drop a field
	int dropAllFields(short fileId);				// drop all on a file

	int findField(short fileId, const char *field);

	int create();						// create a new FIELD.DDF table

	int findId(short Id);

	short getFile() const
		{ return record.Xe_File; }

	short getOffset() const
		{ return record.Xe_Offset; }

	char getDataType() const
		{ return record.Xe_DataType; }

	char getSize() const
		{ return record.Xe_Size; }

	int createSchema(RTdd p);			// sets the database schema

	friend int addField(RTdd dict, short fileId, const char *name, short type, short offset, short len, short dec=0);

	// set the function to call at the bottom of loops.  allows
	// windows to multi-task when the system is busy
	void setLoop(void * obj, void (*func)(void *, int, long))
		{ object = obj; stFunc = func; }
	void yield()
    	{ if(stFunc) stFunc(object, DB_YIELD, 0); }

private:
	int		isOpen;							// true if open
	char 	Xe_Path[MAXPATHLEN+1];			// path name of the FIELD.DDF table
	struct dictField record;				// record read
	char	keyBuf[MAXNAMELEN+sizeof(short)];	// key buffer
	char 	posBlk[ 128 ];					// Btrieve position block
	unsigned short	len;						// length of data read/written
	void (*stFunc)(void *, int, long);
    void *object;
};

struct dictIndex {				// format of the data dictionary INDEX.DDF table
	short	Xi_File;						// file id in dictFile
	short	Xi_Field;						// field if in dictField
	short	Xi_Number;						// index number (0..118)
	short	Xi_Part;						// segment part number (0..118)
	short	Xi_Flags;						// index flags
};

struct cfIndex {							// create structure
	fileSpec	file;
	keySpec		File;						//Xi$File Key
	keySpec		Field;						//Xi$Field Key
	keySpec		File1;
	keySpec		Number;
    keySpec		Part;

    cfIndex();
};

cfIndex::cfIndex() :
	file(sizeof(dictIndex), 512, 2, 0),
	File(0, sizeof(short), Xi_Binary | Xi_Duplicates),
	Field(2, sizeof(short), Xi_Binary | Xi_Duplicates)
	{
	}

class TddIndex {
public:
	TddIndex()
		{ memset(this, 0, sizeof( TddIndex )); }
	TddIndex(const char *path)
		{
		memset(this, 0, sizeof( TddIndex ));
		strncpy(Xi_Path, path, MAXPATHLEN);
		strncat(Xi_Path, "\\INDEX.DDF", MAXPATHLEN);
		}
	~TddIndex()
		{
		}

	void setPath(const char *path)
		{
		strncpy(Xi_Path, path, MAXPATHLEN);
		strncat(Xi_Path, "\\INDEX.DDF", MAXPATHLEN);
		}

	int open()						// open the INDEX.DDF table
		{
		int Result;
		if( isOpen )
        	return btInvalidOperation;

		if( (Result = btOpen(posBlk, Xi_Path))== 0 )
			isOpen = TRUE;

		return Result;
		}

	int close()
	   {
       isOpen = FALSE;
	   return btClose( posBlk );
	   }	

	int dropKey(const Tdd& dict, short indexId);			// drop index on a file

	int getSeg(short file, short index, short part);	// locate a part of an index
    int getFlags(short file, short index, short part);	// get the flags for a segment

	int create();						// create a new INDEX.DDF table

    findId( short Id);

	int createSchema(RTdd p);			// sets the database schema

	friend int addInSeg(RTdd dict, short fileId, short fieldId, short index, short part, short flags);
    friend int addIndex(RTdd dict, short fileId, short index, const char *name);

	// set the function to call at the bottom of loops.  allows
	// windows to multi-task when the system is busy
	void setLoop(void *obj, void (*func)(void *, int, long))
		{ stFunc = func; }
	void yield()
    	{ if(stFunc) stFunc(object, DB_YIELD, 0); }

private:
	int		isOpen;							// true if open
	char 	Xi_Path[MAXPATHLEN+1];			// path name of the INDEX.DDF table
	struct dictIndex record;				// record read
	char	keyBuf[sizeof(short)*3];		// max key size == 6
	char 	posBlk[ 128 ];					// Btrieve position block
	unsigned short	len;						// length of data read/written
	void (*stFunc)(void *, int, long);
    void *object;
};

Tdd::Tdd()
	{
	memset(this, 0, sizeof(this));

	if( (ddFile = new TddFile())== 0 ||
		(ddField = new TddField())== 0 ||
		(ddIndex = new TddIndex())== 0 )
		{
		delete ddFile;
		delete ddField;
		delete ddIndex;

        memset(this, 0, sizeof(this));

		dbError( btMemoryAllocation );
		return;
		}
	}

Tdd::Tdd(const string &path)
	{
    memset(this, 0, sizeof(this));

	if( (ddFile = new TddFile( path.c_str() ))== 0 ||
		(ddField = new TddField( path.c_str() ))== 0 ||
		(ddIndex = new TddIndex( path.c_str() ))== 0 )
		{
		delete ddFile;
		delete ddField;
		delete ddIndex;

        memset(this, 0, sizeof(this));

		dbError( btMemoryAllocation );
		return;
		}
	}

Tdd::~Tdd()
	{
	delete ddFile;
	delete ddField;
	delete ddIndex;
	}

void Tdd::setPath(const string &path)
	{
	ddFile->setPath( path.c_str() );
	ddField->setPath( path.c_str() );
	ddIndex->setPath( path.c_str() );
	}

int Tdd::open()
	{
	int Result;

	(Result = ddFile->open())!= 0 ||
	(Result = ddField->open())!= 0 ||
	(Result = ddIndex->open())!= 0;

	return Result;
	}

int Tdd::close()
	{
	int Result;

	(Result = ddFile->close())!= 0 ||
	(Result = ddField->close())!= 0 ||
	(Result = ddIndex->close())!= 0;

	return Result;
	}

int Tdd::create()
	{
	int Result;

	(Result = ddFile->create())!= 0 ||
	(Result = ddField->create())!= 0 ||
	(Result = ddIndex->create())!= 0;

	if( Result )
		return Result;

	if( (Result = open())!= 0)
    	return Result;

	(Result = ddFile->createSchema( *this ))!= 0 ||
	(Result = ddField->createSchema( *this ))!= 0 ||
	(Result = ddIndex->createSchema( *this ))!= 0;

	if( Result )
    	return Result;

	return close();
	}

// return the upper case table
const char *Tdd::getUprCase() const
	{ return &altSeq[0]; }

void Tdd::setLoop(void *obj, void (*stFunc)(void *, int, long)) const
	{
	ddFile->setLoop(obj, stFunc);
	ddField->setLoop(obj, stFunc);
	ddIndex->setLoop(obj, stFunc);
	}

inline short Tdd::getFileId() const
	{ return getFile().getId(); }

inline short Tdd::getFieldId() const
	{ return getField().getId(); }

int addTable(Tdd& dict, const char *tableName, const char *path, short flags)
	{
	memset(&dict.getFile().record, 0, sizeof( struct dictFile ));	// init all to 0
	dict.getFile().record.Xf_Flags = flags;				// set flags
	strncpy(&dict.getFile().record.Xf_Name[0], tableName, MAXNAMELEN);	// copy the name
	if( strlen(tableName) < MAXNAMELEN )
    	memset(&dict.getFile().record.Xf_Name[strlen(tableName)], ' ', MAXNAMELEN-strlen(tableName));
	strncpy(&dict.getFile().record.Xf_Loc[0], path, MAXPATHLEN);
	if( strlen(path) < MAXPATHLEN )
    	memset(&dict.getFile().record.Xf_Loc[strlen(path)], ' ', MAXPATHLEN-strlen(path));
	dict.getFile().len = sizeof( struct dictFile );
	return BTRCALL(2,
				   (char *)&dict.getFile().posBlk,
				   (char *)&dict.getFile().record,
				   &dict.getFile().len,
				   &dict.getFile().keyBuf[0],
				   MAXNAMELEN,
				   0 );
	}

int addField(Tdd& dict, short fileId, const char *name, short type, short offset, short len, short dec)
	{
	memset(&dict.getField().record, 0, sizeof( struct dictField ));
	dict.getField().record.Xe_File = fileId;
	strncpy(&dict.getField().record.Xe_Name[0], name, MAXNAMELEN);
	if( strlen(name) < MAXNAMELEN )
		memset(&dict.getField().record.Xe_Name[strlen(name)], ' ', MAXNAMELEN-strlen(name));
	dict.getField().record.Xe_DataType = type;
	dict.getField().record.Xe_Offset = offset;
	dict.getField().record.Xe_Size = len;
	dict.getField().record.Xe_Dec = dec;
    dict.getField().len = sizeof( struct dictField );
	return BTRCALL(2,
				   (char *)&dict.getField().posBlk,
				   (char *)&dict.getField().record,
				   &dict.getField().len,	
				   &dict.getField().keyBuf[0],
				   MAXNAMELEN+sizeof(short),
				   0);
	}

int addInSeg(Tdd& dict, short fileId, short fieldId, short index, short part, short flags)
	{
	memset(&dict.getIndex().record, 0, sizeof( struct dictIndex ));
	dict.getIndex().record.Xi_File = fileId;
	dict.getIndex().record.Xi_Field = fieldId;
	dict.getIndex().record.Xi_Number = index;
	dict.getIndex().record.Xi_Part = part;
	dict.getIndex().record.Xi_Flags = flags;
	dict.getIndex().len = sizeof( struct dictIndex );
	return BTRCALL(2,
				   (char *)&dict.getIndex().posBlk,
				   (char *)&dict.getIndex().record,
				   &dict.getIndex().len,
				   &dict.getIndex().keyBuf[0],
				   sizeof(short)*3,
				   0);
	}

int addIndex(Tdd& dict, short fileId, short index, const char *name)
	{
	// and index entry is a field type of 0xFF (all indexes can be named)
    return addField(dict, fileId, name, 0xFF, index, 0, 0);
	}

int TddFile::create()
	{
	if( isOpen )
		return btInvalidOperation;

	cfFile *dbuf = new cfFile;					// create file description
	int Result;

	if( !dbuf )
		return btMemoryAllocation;

	len = sizeof( cfFile );					// size of buffer

	if( (Result = btCreate((void far *)posBlk, (void far *)dbuf, len, Xf_Path, bcCreate))!= 0)
		{
        delete[] dbuf;					// free buffer
		return Result;
		}

	delete[] dbuf;
	return 0;
	}

int TddFile::createSchema(RTdd p)			// sets the database schema
	{
    int Result;

	(Result = addTable(p, "FILE.DDF", Xf_Path, Xf_Dictionary))!= 0 ||	// add record
	(Result = addField(p, p.getFileId(), "Xf$Id", btInteger, 0, 2))!= 0 ||
	(Result = addInSeg(p, p.getFileId(), p.getFieldId(), 0, 0, Xi_Binary))!= 0 ||
	(Result = addField(p, p.getFileId(), "Xf$Name", btString, 2, 20))!= 0 ||
	(Result = addInSeg(p, p.getFileId(), p.getFieldId(), 1, 0, Xi_AltCollateSeq | Xi_Modifiable))!= 0 ||
	(Result = addField(p, p.getFileId(), "Xf$Loc", btString, 22, 64))!= 0 ||
	(Result = addField(p, p.getFileId(), "Xf$Flags", btInteger, 86, 1))!= 0;
	return Result;
	}

int TddFile::dropFile(short id)				// removes a file from the FILE.DDF table
	{
	int Result;

    len = sizeof( dictFile );

	if( (Result = btGetEqual(posBlk,&record, len, &id, 2, 0, btLockWait))!= 0)
		return Result;
	return btDelete(posBlk); 
	}

int TddFile::findFile(const char *name)
	{
	int Result;

	len = sizeof( dictFile );

	char *keyBuf = new char[ MAXNAMELEN ];
    
	if( !keyBuf )
		return 0;					// can't find it

	// build the key
	strncpy(keyBuf, name, MAXNAMELEN);
	if( strlen( name ) < MAXNAMELEN )
		memset( &keyBuf[strlen(name)], ' ', MAXNAMELEN - strlen(name));

	Result = btGetEqual(posBlk, &record, len, keyBuf, MAXNAMELEN, 1);

    delete[] keyBuf;

	if( Result )
		return 0;

	return record.Xf_Id;					// return the Id
	}

int TddFile::findId(short Id)
	{
	if( record.Xf_Id == Id )
		return 0;

	len = sizeof( dictFile );

	return btGetEqual(posBlk, &record, len, (void *)&Id, sizeof(short), 0);
    }

int TddField::create()
	{
	if( isOpen )
		return btInvalidOperation;

	cfField *dbuf = new cfField;					// create file description
	int Result;

	if( !dbuf )
		return btMemoryAllocation;

	len = sizeof( cfField );					// size of buffer

	if( (Result = btCreate((void far *)posBlk, (void far *)dbuf, len, Xe_Path, bcCreate))!= 0)
		{
        delete[] dbuf;					// free buffer
		return Result;
		}

	delete[] dbuf;
	return 0;
	}

class sqlField {							// extended dictionary for SQL
public:
	sqlField()
	: key1(2, 2, Xi_Extended | Xi_Supplemental | Xi_Segment | Xi_Duplicates | Xi_Modifiable, btInteger),
	  key2(25, 2, Xi_Extended | Xi_Supplemental | Xi_Segment | Xi_Duplicates | Xi_Modifiable, btInteger),
	  key3(29, 1, Xi_Extended | Xi_Supplemental | Xi_Duplicates | Xi_Modifiable, btString)
	{
	}

private:
	keySpec	key1, key2, key3;			 	// suplemental key indexes
};

int TddField::createSchema(RTdd p)			// sets the database schema
	{
	short file, name, offset, dec;
    int Result;

	(Result = addTable(p, "FIELD.DDF", Xe_Path, Xf_Dictionary))!= 0 ||	// add record
	(Result = addField(p, p.getFileId(), "Xe$Id", btInteger, 0, 2))!= 0 ||
	(Result = addInSeg(p, p.getFileId(), p.getFieldId(), 0, 0, Xi_Binary))!= 0 ||
	(Result = addField(p, p.getFileId(), "Xe$File", btInteger, 2, 2))!= 0 ||
	(Result = addInSeg(p, p.getFileId(), file = p.getFieldId(), 1, 0, Xi_Duplicates | Xi_Binary))!= 0 ||
	(Result = addField(p, p.getFileId(), "Xe$Name", btString, 4, 20))!= 0 ||
	(Result = addInSeg(p, p.getFileId(), name = p.getFieldId(), 2, 0, Xi_Duplicates | Xi_AltCollateSeq | Xi_Modifiable))!= 0 ||
	(Result = addField(p, p.getFileId(), "Xe$DataType", btInteger, 24, 1))!= 0 ||
	(Result = addField(p, p.getFileId(), "Xe$Offset", btInteger, 25, 2))!= 0 ||
    ((offset = p.getFieldId()),
	(Result = addField(p, p.getFileId(), "Xe$Size", btInteger, 27, 2)))!= 0 ||
	(Result = addField(p, p.getFileId(), "Xe$Dec", btInteger, 29, 1))!= 0 ||
    ((dec = p.getFieldId()),
	(Result = addField(p, p.getFileId(), "Xe$Flags", btInteger, 30, 2)))!= 0 ||
	(Result = addInSeg(p, p.getFileId(), file, 3, 0, Xi_Binary | Xi_Segment | Xi_Modifiable))!= 0 ||
	(Result = addInSeg(p, p.getFileId(), name, 3, 1, Xi_AltCollateSeq | Xi_Modifiable))!= 0;

	if( Result )
    	return Result;

	// create the SQL extended indexes
	sqlField	supIndex;

	(Result = addInSeg(p, p.getFileId(), file, 4, 0, Xi_Extended | Xi_Supplemental | Xi_Segment | Xi_Duplicates | Xi_Modifiable))!= 0 ||
	(Result = addInSeg(p, p.getFileId(), offset, 4, 1, Xi_Extended | Xi_Supplemental | Xi_Segment | Xi_Duplicates | Xi_Modifiable))!= 0 ||
	(Result = addInSeg(p, p.getFileId(), dec, 4, 2, Xi_Extended | Xi_Supplemental |  Xi_Duplicates | Xi_Modifiable))!= 0;

	if( Result )
    	return Result;

   	return btCreateSupplementalIndex(&p.getField().posBlk, &supIndex, sizeof( supIndex ));
	}

int TddField::dropField(short field)
	{
	int Result;

    len = sizeof( dictField );

	if( (Result = btGetEqual(posBlk, &record, len, &field, 2, 0, btLockWait))!= 0)
		return Result;
	return btDelete(posBlk); 
	}

int TddField::dropAllFields(short file)
	{
	int Result;

    len = sizeof( dictField );

	if( (Result = btGetEqual(posBlk, &record, len, &file, 2, 1, btLockWait))!= 0)
		{
		switch( Result )
			{
			case btKeyValueNotFound :
				return 0;
			default :
				return Result;
			}
		}

    // found at least 1 field so drop the rest
	while( 1 )
		{
		if( (Result = btDelete(posBlk))!= 0)
			return Result;						// failed

		len = sizeof( dictField );

		if( (Result = btGetNext(posBlk, &record, len, &file, 2, 1, btLockWait))!= 0)
			{
			switch( Result )
            	{
				case btEndOfFile :
					return 0;
				default :
					return Result;
                }
			}
        yield();
		}
    }

int TddField::findField(short file, const char *name)
	{
	int Result;

	len = sizeof( dictField );

	char *keyBuf = new char[ sizeof( short ) + MAXNAMELEN ];
    
	if( !keyBuf )
		return 0;					// can't find it

	// build the key
	*((short *)keyBuf) = file;		// set the file Id
	strncpy(&keyBuf[sizeof(short)], name, MAXNAMELEN);
	if( strlen(name) < MAXNAMELEN )
		memset( &keyBuf[sizeof(short)+strlen(name)], ' ', MAXNAMELEN-strlen(name));

	Result = btGetEqual(posBlk, &record, len, keyBuf, sizeof(short)+MAXNAMELEN, 3);

    delete[] keyBuf;

	if( Result )
		return 0;

	return record.Xe_Id;					// return the Id
    }

int TddField::findId(short Id)
	{
	if( record.Xe_Id == Id )
		return 0;

	len = sizeof( dictField );

	return btGetEqual(posBlk, &record, len, (void *)&Id, sizeof(short), 0);
    }

int TddIndex::create()
	{
	if( isOpen )
		return btInvalidOperation;

	cfIndex *dbuf = new cfIndex;					// create file description
	int Result;

	if( !dbuf )
		return btMemoryAllocation;

	len = sizeof( cfIndex );					// size of buffer

	if( (Result = btCreate((void far *)posBlk, (void far *)dbuf, len, Xi_Path, bcCreate))!= 0)
		{
        delete[] dbuf;					// free buffer
		return Result;
		}

	delete[] dbuf;
	return 0;
	}

class sqlIndex {
public:
	sqlIndex()
	: key1(0, 2, Xi_Extended | Xi_Supplemental | Xi_Segment | Xi_Modifiable, btInteger),
	  key2(4, 2, Xi_Extended | Xi_Supplemental | Xi_Segment | Xi_Modifiable, btInteger),
	  key3(6, 2, Xi_Extended | Xi_Supplemental | Xi_Modifiable, btInteger)
	{
	}

private:
	keySpec	key1, key2, key3;
};

int TddIndex::createSchema(RTdd p)			// sets the database schema
	{
    short file, number, part;
	int Result;

	(Result = addTable(p, "INDEX.DDF", Xi_Path, Xf_Dictionary))!= 0 ||	// add record
	(Result = addField(p, p.getFileId(), "Xi$File", btInteger, 0, 2))!= 0 ||
	(Result = addInSeg(p, p.getFileId(), file = p.getFieldId(), 0, 0, Xi_Duplicates | Xi_Extended))!= 0 ||
	(Result = addField(p, p.getFileId(), "Xi$Field", btString, 2, 2))!= 0 ||
	(Result = addInSeg(p, p.getFileId(), p.getFieldId(), 1, 0, Xi_Duplicates | Xi_Extended))!= 0 ||
	(Result = addField(p, p.getFileId(), "Xi$Number", btInteger, 4, 2))!= 0 ||
    ((number = p.getFieldId()),
	(Result = addField(p, p.getFileId(), "Xi$Part", btInteger, 6, 2)))!= 0 ||
    ((part = p.getFieldId()),
	(Result = addField(p, p.getFileId(), "Xi$Flags", btInteger, 8, 2)))!= 0;

	if( Result )
		return Result;

	// now create the SQL supplemental index needed
	sqlIndex	supIndex;

	(Result = addInSeg(p, p.getFileId(), file, 2, 0, Xi_Extended | Xi_Supplemental | Xi_Segment | Xi_Modifiable))!= 0 ||
	(Result = addInSeg(p, p.getFileId(), number, 2, 1, Xi_Extended | Xi_Supplemental | Xi_Segment | Xi_Modifiable))!= 0 ||
	(Result = addInSeg(p, p.getFileId(), part, 2, 2, Xi_Extended | Xi_Supplemental | Xi_Modifiable))!= 0;

	if( Result )
		return Result;

    return btCreateSupplementalIndex(&p.getIndex().posBlk, &supIndex, sizeof( supIndex ));
	}

int TddIndex::dropKey(const Tdd& p, short index)
	{
	int Result;
    short	indexId;
    short	file;

    // locate the index in the extended dictionary file
	if( (Result = p.getField().findId(index))!= 0)
		return Result;

    indexId = p.getField().getId();
    file = p.getField().getFile();

	// now look up the index in the standard block
	len = sizeof( dictIndex );

	if( (Result = btGetEqual(posBlk, &record, len, &file, 2, 0, btLockWait))== 0)
		{
		while( 1 )						// search for all keys
			{
			if( record.Xi_Number == indexId )
				if( (Result = btDelete(posBlk))!= 0)
					return Result;

			len = sizeof( dictIndex );
			if( (Result = btGetNext(posBlk, &record, len, &file, 2, 0, btLockWait))!= 0)
				{
				if( Result == btEndOfFile )
					break;
				return Result;
				}
            yield();
            }
		}

	if( Result != btKeyValueNotFound &&
		Result != btEndOfFile )
		return Result;

	// now delete the named entry
	return p.getField().dropField(index);
	}

// locate a part of an index.  returns the field Id of the part
int TddIndex::getSeg(short file, short index, short seg)
	{
	if( record.Xi_File == file &&
		record.Xi_Number == index &&
		record.Xi_Part == seg )
        	return record.Xi_Field;

	short key[3];						// build the key here

	key[0] = file;
	key[1] = index;
	key[2] = seg;

	len = sizeof( dictIndex );
	if( btGetEqual(posBlk, &record, len, key, sizeof( key ), 2)!= 0)
		return 0;						// not found

	return record.Xi_Field;				// return the field number
	} 

// locate the field flags for a segment of an index, (must exist!)
int TddIndex::getFlags(short file, short index, short seg)
	{
	if( record.Xi_File == file &&
		record.Xi_Number == index &&
		record.Xi_Part == seg )
        	return record.Xi_Flags;

	short key[3];						// build the key here

	key[0] = file;
	key[1] = index;
	key[2] = seg;

	len = sizeof( dictIndex );
	if( btGetEqual(posBlk, &record, len, key, sizeof( key ), 2)!= 0)
		return 0;						// not found

	return record.Xi_Flags;				// return the field number
	} 

int Tdd::addTable(const string &name, const string &path)
	{
	return ::addTable(*this, name.c_str(), path.c_str(), 0);
	}

int Tdd::dropTable(short Id) const
	{
	return getFile().dropFile( Id );
	}

int Tdd::getTableId(const string &name) const
	{
	return getFile().findFile( name.c_str() );
	}

const char *Tdd::getTablePath(short Id) const
	{
	if( getFile().findId( Id )!= 0)
		return "";

	return getFile().getLoc();
    }

int Tdd::addField(short file, const string &name, char type, short offset, short size, char dec)
	{
	return ::addField(*this, file, name.c_str(), type, offset, size, dec);
	}

int Tdd::dropField(short field) const
	{
	return getField().dropField(field);
	}

int Tdd::dropAllFields(short file) const
	{
	return getField().dropAllFields(file);
	}

int Tdd::getFieldId(short file, const string &name) const
	{
	return getField().findField(file, name.c_str());
	}

int Tdd::getFieldOffset(short Id) const
	{
	if( getField().findId( Id )!= 0)
    	return 0;

	return getField().getOffset();
	}

int Tdd::getFieldLength(short Id) const
	{
	if( getField().findId( Id )!= 0)
		return 0;

    return getField().getSize();
	}

int Tdd::getFieldType(short Id) const
	{
	if( getField().findId( Id )!= 0)
    	return 0;

	return getField().getDataType();
    }

int Tdd::addSeg(short file, short field, short num, short seq, short flags)
	{
	return ::addInSeg(*this, file, field, num, seq, flags);
	}

int Tdd::addKey(short file, short num, const string &name)
	{
	return ::addIndex(*this, file, num, name.c_str());
	}

int Tdd::dropKey(short index) const
	{
	return getIndex().dropKey(*this, index );
	}

int Tdd::getKeyId(short file, const string &name) const
	{
	return getField().findField(file, name.c_str());
	}

int Tdd::getKeyNum(short Id) const
	{
	if( getField().findId( Id )!= 0)
    	return 0;

	return getField().getOffset();
    }

int Tdd::getSeg(short file, short index, short seg) const
	{
	return getIndex().getSeg(file, index, seg);
	}

int Tdd::getFlags(short file, short index, short seg) const
	{
	return getIndex().getFlags(file, index, seg);
	}
